/******************************************************************************
 * (c) Copyright 2002-2006, 1060 Research Ltd
 *
 * This Software is licensed to You, the licensee, for use under the terms of
 * the 1060 Public License v1.0. Please read and agree to the 1060 Public
 * License v1.0 [www.1060research.com/license] before using or redistributing
 * this software.
 *
 * In summary the 1060 Public license has the following conditions.
 * A. You may use the Software free of charge provided you agree to the terms
 * laid out in the 1060 Public License v1.0
 * B. You are only permitted to use the Software with components or applications
 * that provide you with OSI Certified Open Source Code [www.opensource.org], or
 * for which licensing has been approved by 1060 Research Limited.
 * You may write your own software for execution by this Software provided any
 * distribution of your software with this Software complies with terms set out
 * in section 2 of the 1060 Public License v1.0
 * C. You may redistribute the Software provided you comply with the terms of
 * the 1060 Public License v1.0 and that no warranty is implied or given.
 * D. If you find you are unable to comply with this license you may seek to
 * obtain an alternative license from 1060 Research Limited by contacting
 * license@1060research.com or by visiting www.1060research.com
 *
 * NO WARRANTY:  THIS SOFTWARE IS NOT COVERED BY ANY WARRANTY. SEE 1060 PUBLIC
 * LICENSE V1.0 FOR DETAILS
 *
 * THIS COPYRIGHT NOTICE IS *NOT* THE 1060 PUBLIC LICENSE v1.0. PLEASE READ
 * THE DISTRIBUTED 1060_Public_License.txt OR www.1060research.com/license
 *
 * File:          $RCSfile$
 * Version:       $Name$ $Revision$
 * Last Modified: $Date$
 *****************************************************************************/
package org.netkernel.layer0.nkf.impl;


import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.urii.EndpointImpl;
import org.netkernel.layer0.util.Utils;
import org.netkernel.request.*;
import org.netkernel.request.impl.ExpiryFactory;
import org.netkernel.request.impl.MetaImpl;
import org.netkernel.request.impl.RequestFactory;
import org.netkernel.request.impl.ResponseImpl;
import org.netkernel.scheduler.ResolutionFailureException;
import org.netkernel.urii.*;

/**
 * The main base-class for creating accessors using the NetKernel Foundation API
 * @author  tab
 */
public abstract class NKFEndpointImpl extends EndpointImpl implements IEndpoint 
{
	
	public static final boolean SAFE_FOR_CONCURRENT_USE=true;
	public static final boolean NOT_SAFE_FOR_CONCURRENT_USE=false;
	
	private INKFParametersControl mParameters;
	
    /** Creates a new instance of NKFAccessorImpl */
    public NKFEndpointImpl()
    {	super();
    }
	
    /** Creates a new instance of NKFAccessorImpl */
    public NKFEndpointImpl(boolean aIsThreadSafe)
    {	super(aIsThreadSafe);
    }
    
    protected void setParameters(INKFParametersControl aParameters)
    {	mParameters=aParameters;
    }
    
    protected INKFParametersControl getParameters()
    {	return mParameters;
    }
    
	/** subclasses of this accessor cannot issue any asychronous requests */
	//public final void receiveAsyncException(IResponse aResult) {}
	/** subclasses of this accessor cannot issue any asychronous requests */
	//public final void receiveAsyncResult(IResponse aResult) {}

	public void onAsyncRequest(IRequest aRequest)
    {	IResponse response;
		try
		{	if (getKernel()==null)
			{	// can't use generated message here because don't have a handle to kernel!
				throw new NKFException("Endpoint not commissioned",this.getClass().getName());
			}
		
			NKFRequestReadOnlyImpl req=createThisNKFRequest(aRequest);
			NKFContextImpl context=new NKFContextImpl(getKernel(),getSpace(),req,this,mParameters);
			req.setContext(context);
			onRequest(context);
			response = context.getKernelResponse();
			context.destroy();
		}
		catch (InterruptedException e)
		{	response = new ResponseImpl(aRequest,e,new MetaImpl(ExpiryFactory.getAlwaysExpired()));
		}
		catch (Exception e)
		{	Throwable t=e;
			boolean rfe=false;
			if (e instanceof NKFException)
			{	NKFException nke=(NKFException)e;
				if (areUnhandledExceptionsVerbose())
				{	if (nke.getLocation()==null)
					{	nke.setShowNativeStack();
					}
				}
				else
				{	if (nke.getId().equals(NKFException.class.getName()))
					{	t=nke.getCause();
					}
				}
				Throwable dc=nke.getDeepestCause();
				if (dc!=null && dc instanceof ResolutionFailureException)
				{	// resolution failures are immune from being cached
					rfe=true;
				}
			}
			
			IExpiry expiry = (areUnhandledExceptionsExpired()||rfe)?ExpiryFactory.getAlwaysExpired():ExpiryFactory.getDependentExpiry();
			response = new ResponseImpl(aRequest,t,new MetaImpl(expiry));
		}
		if (response!=null)
		{	aRequest.getRequestor().onResponse(response);
		}
    }
	
	protected NKFRequestReadOnlyImpl createThisNKFRequest(IRequest aRequest) throws Exception
	{	NKFRequestReadOnlyImpl result=null;
		if (aRequest!=null)
		{	result = new NKFRequestReadOnlyImpl(aRequest);
		}
		return result;
	}
	
	/** subclasses can override this method to have access to all verbs in one method
	 * alternatively use the on[Verb] methods for finer grained control
	 * @arg aContext the NKF context object
	 * @throws Throwable if anything goes wrong
	 */
	public void onRequest(INKFRequestContext aContext) throws Exception
	{
		switch (aContext.getKernelContext().getThisKernelRequest().getVerb())
		{	
			case INKFRequestReadOnly.VERB_SOURCE:
				onSource(aContext);
				break;
			case INKFRequestReadOnly.VERB_SINK:
				onSink(aContext);
				break;
			case INKFRequestReadOnly.VERB_EXISTS:
				onExists(aContext);
				break;
			case INKFRequestReadOnly.VERB_DELETE:
				onDelete(aContext);
				break;
			case INKFRequestReadOnly.VERB_NEW:
				onNew(aContext);
				break;
			case INKFRequestReadOnly.VERB_TRANSREPRESENT:
				onTransrept(aContext);
				break;
			case INKFRequestReadOnly.VERB_META:
				onMeta(aContext);
				break;
			default:
				unsupportedVerb(aContext);
				break;
		}
	}
	
	public void onSource(INKFRequestContext aContext) throws Exception
	{	unsupportedVerb(aContext);
	}
	public void onSink(INKFRequestContext aContext) throws Exception
	{	unsupportedVerb(aContext);
	}
	public void onExists(INKFRequestContext aContext) throws Exception
	{	unsupportedVerb(aContext);
	}
	public void onDelete(INKFRequestContext aContext) throws Exception
	{	unsupportedVerb(aContext);
	}
	public void onNew(INKFRequestContext aContext) throws Exception
	{	unsupportedVerb(aContext);
	}
	public void onTransrept(INKFRequestContext aContext) throws Exception
	{	unsupportedVerb(aContext);
	}
	public void onMeta(INKFRequestContext aContext) throws Exception
	{	unsupportedVerb(aContext);
	}

	protected void unsupportedVerb(INKFRequestContext aContext) throws Exception
	{	String endpointClass = this.getClass().getName();
		String verb=RequestFactory.verbString(aContext.getThisRequest().getVerb());
		throw Utils.createFormattedException("EX_NKF_VERB_NOT_SUPPORTED", "MSG_NKF_VERB_NOT_SUPPORTED", aContext.getKernelContext().getKernel().getLogger(), verb,endpointClass);
	}

}