package org.netkernel.layer0.representation.impl;

import java.util.ArrayList;
import java.util.List;

import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.representation.IHDSPredicate;
import org.netkernel.layer0.util.HDSXPath;

public class HDSNodeImpl implements IHDSNode
{
	private final String mName;
	private final IHDSNode mParent;
	private Object mValue;
	private IHDSNode[] mChildren;
	private int mDeepHash;
	
	@Override
	public int hashCode()
	{	return mDeepHash;
	}

	@Override
	public boolean equals(Object aOther)
	{	boolean result=this==aOther;
		if (!result)
		{	if (aOther instanceof HDSNodeImpl)
			{	HDSNodeImpl other=(HDSNodeImpl)aOther;
				result=mDeepHash==other.mDeepHash;
				if (result)
				{	result=((mName==other.mName) || (mName!=null && mName.equals(other.mName))) && ((mValue==other.mValue) || (mValue!=null && mValue.equals(other.mValue))) && mChildren.length==other.mChildren.length;
					if (result)
					{	for (int i=0; i<mChildren.length && result; i++)
						{	result=mChildren[i].equals(other.mChildren[i]);
						}
					}
				}
			}
			else
			{	result=false;
			}
		}
		return result;
	}
	
	private void setDirty()
	{
		int hash=0;
		if (mName!=null)
		{	hash=mName.hashCode();
		}
		if (mValue!=null)
		{	hash^=mValue.hashCode();
		}
		for (int i=0; i<mChildren.length; i++)
		{	hash^=mChildren[i].hashCode();
		}
		mDeepHash=hash;
	}

	private static IHDSNode[] NO_CHILDREN=new IHDSNode[0];
	
	public HDSNodeImpl(String aName, Object aValue, IHDSNode aParent)
	{	mName=aName;
		mValue=aValue;
		mParent=aParent;
		mChildren=NO_CHILDREN;
		setDirty();
	}
	
	public HDSNodeImpl(String aName, IHDSNode aParent)
	{	this(aName,null,aParent);
		setDirty();
	}
	
	public void setChildren(IHDSNode[] aChildren)
	{	mChildren=aChildren;
		setDirty();
	}
	
	public void setChildren(List<IHDSNode> aChildren)
	{	mChildren=new IHDSNode[aChildren.size()];
		mChildren=aChildren.toArray(mChildren);
		setDirty();
	}
	
	public void setValue(Object aValue)
	{	mValue=aValue;
		setDirty();
	}

	public String getName()
	{	return mName;
	}

	public IHDSNode[] getChildren()
	{	return mChildren;
	}

	public IHDSNodeList getNodes(String aXPath)
	{	List<IHDSNode> nodes=new ArrayList<IHDSNode>();
		HDSXPath.eval(nodes,this, aXPath);
		return new HDSNodeListImpl(nodes);
	}

	public Object getValue()
	{	return mValue;
	}

	public Object[] getValues(String aXPath)
	{	List<IHDSNode> nodes=new ArrayList<IHDSNode>();
		HDSXPath.eval(nodes,this, aXPath);
		int l=nodes.size();
		Object[] result = new Object[l];
		for (int i=0; i<l; i++)
		{	result[i]=((IHDSNode)nodes.get(i)).getValue();
		}
		return result;
	}

	public IHDSNode getParent()
	{	return mParent;
	}

	public IHDSNode getNamedChild(String name)
	{	for(IHDSNode node: mChildren)
		{	if(node.getName().equals(name))
			{	return node;
			}
		}
		return null;
	}

	public IHDSNode getRoot()
	{	IHDSNode root=this;
		IHDSNode parent=null;
		while(true)
		{	parent=root.getParent();
			if(parent!=null)
			{		root=parent;
			}
			else break;
		}
		return root;
	}
	
	public String toString()
	{	StringBuffer sb=new StringBuffer();
		recurseToString(sb, this,"");	
		return sb.toString();
	}
	
	private void recurseToString(StringBuffer sb, IHDSNode n, String indent)
	{	sb.append(indent);
		sb.append(n.getName());
		sb.append(" : ");
		if(n.getValue()!=null)
		{	sb.append(n.getValue().toString());
		}
		else
		{	sb.append("null");
		}
		sb.append("\n");
		IHDSNode[] children=n.getChildren();
		if(children!=null)
		{	indent=indent+"\t";
			for(int i=0; i< children.length; i++)
			{	recurseToString(sb, children[i], indent);
			}
		}
	}

	public IHDSNode getFirstNode(String aXPath)
	{	return HDSXPath.getSingleNode(this, aXPath,false);
	}

	public Object getFirstValue(String aXPath)
	{	IHDSNode node = HDSXPath.getSingleNode(this, aXPath,false);
		return (node!=null)?node.getValue():null;
	}

	public IHDSNode[] getChildren(IHDSPredicate aPredicate)
	{	IHDSNode[] result = new IHDSNode[mChildren.length];
		int i=0;
		for (IHDSNode child: mChildren)
		{	if (aPredicate.matches(child))
			{	result[i++]=child;
			}
		}
		if (i!=result.length)
		{	if (i==0)
			{	result = NO_CHILDREN;
			}
			else
			{	IHDSNode[] result2 = new IHDSNode[i];
				System.arraycopy(result, 0, result2, 0, i);
				result=result2;
			}
		}
		return result;
	}

	public IHDSNode getFirstChild(IHDSPredicate aPredicate)
	{	for (IHDSNode child: mChildren)
		{	if (aPredicate.matches(child))
			{	return child;
			}
		}
		return null;
	}

	public int getPositionInParent()
	{	IHDSNode[] siblings = mParent.getChildren();
		for (int i=0; i<siblings.length; i++)
		{	if (siblings[i]==this)
			{	return i;
			}
		}
		throw new IllegalArgumentException("IHDSNode corrupt");
	}

	public String getXPath()
	{	StringBuilder b = new StringBuilder();
		IHDSNode n = this;
		List ancestors = new ArrayList();
		do
		{	ancestors.add(n);
			n=n.getParent();
		} while (n!=null);
		if (ancestors.size()>1)
		{
			for (int i=ancestors.size()-2; i>=0; i--)
			{	b.append('/');
				n=(IHDSNode)ancestors.get(i);
				String name=n.getName();
				b.append(name);
				IHDSNode[] siblings=n.getParent().getChildren();
				int countInParent=0;
				int indexInParent=0;
				for (int j=0; j<siblings.length; j++)
				{	if (siblings[j].getName()==name)
					{	if (siblings[j]==n)
						{	indexInParent=j+1;
						}
						countInParent++;
					}
				}
				if (countInParent>1)
				{	b.append('[');
					b.append(Integer.toString(indexInParent));
					b.append(']');
				}
			}
		}
		else
		{	b.append('/');
		}
		return b.toString();
	}
}
